#!/bin/bash

# Simple version updater working on the Version Header file.
# Updates build, minor or major version tag.


APPLICATION=$1
MODE=$2


if [ "$APPLICATION" == "" ]
then
  echo
  echo "Usage: $0 {ApplicationName} [BUILD|MINOR|MAJOR|"
  echo
  echo "BUILD:"
  echo "The build version number in the Version Header File {ApplicationName}.ver"
  echo "will be incremented by one."
  echo
  echo "MINOR:"
  echo "The minor version number in the Version Header File {ApplicationName}.ver"
  echo "will be incremented by one."
  echo
  echo "MAJOR:"
  echo "The major version number in the Version Header File {ApplicationName}.ver"
  echo "will be incremented by one. The minor number is reset to zero."
  echo
  echo "If second argument is not given, BUILD will be used."
  echo
  exit 1
fi

if [ "$MODE" == "" ]
then
  MODE=BUILD
fi



TEMP_FILE=${APPLICATION}.tmp
VER_FILE=${APPLICATION}.ver

NUMBER=`grep "VERSION_$MODE" $VER_FILE | sed "s/.\+[ \t]\([0-9]\+\)/\1/"`
export NEW_NUMBER=$(( NUMBER + 1 ))

echo "Updating $MODE from $NUMBER to $NEW_NUMBER in $VER_FILE"



# Step 1: Update the Version Header File

if [ $MODE == BUILD ] ; then
  gawk '/VERSION_BUILD/ { print $1 " " $2 " " ENVIRON["NEW_NUMBER"] ; next } { print $0 }' $VER_FILE > $TEMP_FILE
fi

if [ $MODE == MINOR ] ; then
  gawk '/VERSION_MINOR/ { print $1 " " $2 " " ENVIRON["NEW_NUMBER"] ; next } { print $0 }' $VER_FILE > $TEMP_FILE
fi

if [ $MODE == MAJOR ] ; then
  gawk '/VERSION_MAJOR/ { print $1 " " $2 " " ENVIRON["NEW_NUMBER"] ; next } /VERSION_MINOR/ { print $1 " " $2 " 0" ; next } { print $0 }' $VER_FILE > $TEMP_FILE
fi

if [ $? != 0 ]
then
  echo "Failed updating Version Header File $VER_FILE"
  exit 10
fi

chmod +w $VER_FILE
cp -f $TEMP_FILE $VER_FILE
rm $TEMP_FILE

exit 0
