
#include "SidPlayDll.h"

#include <E32Math.h>
#include <StdLib.h>

#include <include/emucfg.h>
#include <include/sidtune.h>

#include <sound/SoundParameters.h>



GLREF_C bool sidEmuInitializeSong(emuEngine& thisEmuEngine, sidTune& thisTune, uword songNumber);



enum TSidPlayDllErrors
{
	KErrEndianess = 10,
	KErrNoStereoPossible,
	KErrSidTuneBadStatus,
	KErrSidEmuBadSettings,
	KErrSidEmuInitSong
};



EXPORT_C CSidPlayDll * CSidPlayDll::NewL()
{
    CSidPlayDll * self = NewLC();
	CleanupStack::Pop( self );
	return self;
}



EXPORT_C CSidPlayDll * CSidPlayDll::NewLC()
{
	CSidPlayDll * self = new ( ELeave ) CSidPlayDll();
	CleanupStack::PushL( self );
	self->ConstructL();
	return self;
}



CSidPlayDll::CSidPlayDll()
{

}



CSidPlayDll::~CSidPlayDll()
{
	Stop();
	Unload();

	delete iEmuEngine; iEmuEngine = NULL;
	delete iTune; iTune = NULL;
	if ( iEmuConfig ) free( iEmuConfig ); iEmuConfig = NULL;
	if ( iSidInfo ) free( iSidInfo ); iSidInfo = NULL;
}



void CSidPlayDll::ConstructL()
{
	iPlaying = EFalse;
	iPaused = EFalse;

	iPlayFrequency = 16000;
	iStereo = EFalse;

	iEmuConfig = ( struct emuConfig* ) malloc( sizeof( struct emuConfig ) );
	Mem::FillZ( iEmuConfig, sizeof( struct emuConfig ) );

	iSidInfo = ( struct sidTuneInfo* ) malloc( sizeof( struct sidTuneInfo ) );
	Mem::FillZ( iSidInfo, sizeof( struct sidTuneInfo ) );

	iEmuEngine = new ( ELeave ) emuEngine();
	if ( iEmuEngine->verifyEndianess() == false )
	{
		User::Leave( KErrEndianess );
	}

	// Get the default configuration
	iEmuEngine->getConfig( *iEmuConfig );

	// Validate emulator settings
	if ( ( iEmuConfig->autoPanning != SIDEMU_NONE ) && ( iEmuConfig->channels == SIDEMU_MONO ) )
	{
		User::Leave( KErrNoStereoPossible );
	}
	if ( ( iEmuConfig->autoPanning != SIDEMU_NONE ) && ( iEmuConfig->volumeControl == SIDEMU_NONE ) )
	{
		iEmuConfig->volumeControl = SIDEMU_FULLPANNING;
	}

	UpdateWaveFormatL();
}



EXPORT_C void CSidPlayDll::LoadL( const TDesC8 & aModuleData )
{
	Unload();

	// Instantiate a sidTune object
	iTune = new ( ELeave ) sidTune( ( ubyte * ) aModuleData.Ptr(), ( udword ) aModuleData.Length() );
	iTune->getInfo( *iSidInfo );
	if ( !iTune )	// Call of the fucking stupid () operator?
	{
		User::Leave( KErrSidTuneBadStatus );
	}

	iTune->getInfo( *iSidInfo );
	if ( !iTune )	// Call of the fucking stupid () operator?
	{
		User::Leave( KErrSidTuneBadStatus );
	}

	if ( sidEmuInitializeSong( *iEmuEngine, *iTune, iSidInfo->startSong ) == false )
	{
		User::Leave( KErrSidEmuInitSong );
	}
}



EXPORT_C void CSidPlayDll::Unload()
{
	Stop();
	delete iTune; iTune = NULL;
}



EXPORT_C void CSidPlayDll::Start()
{
	if ( !iPlaying && iTune != NULL )
	{
		iPlaying = ETrue;
	}
}



EXPORT_C void CSidPlayDll::Stop()
{
	if ( iPlaying )
	{
		if ( iTune != NULL )
		{
			sidEmuInitializeSong( *iEmuEngine, *iTune, 0 );
		}
		iPlaying = EFalse;
	}
}



EXPORT_C void CSidPlayDll::Pause( TBool aPauseFlag )
{
	iPaused = aPauseFlag;
}



void CSidPlayDll::InitL( MSoundParameters * aSoundParamaters )
{
    SetPlayFrequencyL( aSoundParamaters->SampleRate() );
}



void CSidPlayDll::AddTo( TDes16 & aSoundBuffer )
{
	if ( iPlaying && !iPaused && iTune != NULL )
	{
		iEmuEngine->FillBuffer( *iTune, ( void * ) aSoundBuffer.Ptr(), ( udword ) ( aSoundBuffer.Length() * 2 ) );
	}
}



EXPORT_C void CSidPlayDll::SetVolume( TUint8 /*aVolume*/ )
{

}



EXPORT_C TUint8 CSidPlayDll::Volume()
{
    return MaxVolume();
}



EXPORT_C TUint8 CSidPlayDll::MaxVolume()
{
	return 255;
}



void CSidPlayDll::UpdateWaveFormatL()
{
	TBool playing = iPlaying;
	if ( playing )
	{
		iPlaying = EFalse;
	}

	iEmuConfig->frequency = ( uword ) iPlayFrequency;
	iEmuConfig->channels = iStereo ? SIDEMU_STEREO : SIDEMU_MONO;
	iEmuConfig->bitsPerSample = SIDEMU_16BIT;

	if( iEmuEngine->setConfig( *iEmuConfig ) == false )
	{
		User::Leave( KErrSidEmuBadSettings );
	}

	if ( playing )
	{
		iPlaying = ETrue;
	}
}



void CSidPlayDll::SetStereoL( TBool aStereoFlag )
{
	if ( aStereoFlag != iStereo )
	{
		iStereo = aStereoFlag;
		UpdateWaveFormatL();
	}
}



void CSidPlayDll::SetPlayFrequencyL( TInt aFrequency )
{
	if ( aFrequency != iPlayFrequency )
	{
		switch ( aFrequency )
		{
		case 44100:
		case 22050:
		case 16000:
		case 11025:
			iPlayFrequency = aFrequency;
			UpdateWaveFormatL();
			break;

		default:
			User::Leave( KErrArgument );
		}
	}
}



GLDEF_C TInt E32Dll( TDllReason /*aReason*/ )
{
    return KErrNone;
}



GLDEF_C TInt ftol2( float aFloat )
{
	TInt32 int32 = 0;
	TInt error = Math::Int( int32, aFloat + 0.5f );
	if ( error != KErrNone )
	{
		User::Panic( _L( "__ftol2" ), error );
	}
	return ( TInt ) int32;
}
